"""Before Data"""

from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from typing import List

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike, NDArray

from ._epochs import BeForEpochs

pd.set_option("mode.copy_on_write", True)


@dataclass
class BeForRecord:
    """Data Structure for handling behavioural force measurements

    Args
    ----
    dat: Pandas Dataframe
        data
    sampling_rate: float
        the sampling rate of the force measurements
    sessions: list of integer
        sample numbers at which a new recording session starts, if the exists
    time_column :
            str = ""
    meta: dictionary
        any kind of meta data
    """

    dat: pd.DataFrame
    sampling_rate: float
    sessions: List[int] = field(default_factory=list[int])
    time_column: str = ""
    meta: dict = field(default_factory=dict)

    def __post_init__(self):
        if not isinstance(self.dat, pd.DataFrame):
            raise TypeError(f"must be pandas.DataFrame, not {type(self.dat)}")

        if len(self.sessions) == 0:
            self.sessions.append(0)
        else:
            if isinstance(self.sessions, int):
                self.sessions = [self.sessions]
            if self.sessions[0] < 0:
                self.sessions[0] = 0
            elif self.sessions[0] > 0:
                self.sessions.insert(0, 0)

        if len(self.time_column) > 0 and self.time_column not in self.dat:
            raise ValueError(f"Time column {self.time_column} not found in DataFrame")

    def __repr__(self):
        rtn = "BeForRecord"
        rtn += f"\n  sampling_rate: {self.sampling_rate}"
        rtn += f", n sessions: {self.n_sessions()}"
        if len(self.time_column) >= 0:
            rtn += f"\n  time_column: {self.time_column}"
        rtn += "\n  metadata"
        for k, v in self.meta.items():
            rtn += f"\n  - {k}: {v}".rstrip()
        rtn += "\n" + str(self.dat)
        return rtn

    def force_cols(self) -> NDArray[np.intp]:
        """List of indices of the columns that contain the force data"""
        return np.flatnonzero(self.dat.columns != self.time_column)

    def n_samples(self) -> int:
        """Number of sample in all sessions"""
        return self.dat.shape[0]

    def n_forces(self) -> int:
        """Number of force columns"""
        return len(self.force_cols())

    def n_sessions(self) -> int:
        """Number of recoding sessions"""
        return len(self.sessions)

    def time_stamps(self) -> NDArray:
        """The time stamps (numpy array)

        Creates time stamps, of they are not define in the data
        """
        if len(self.time_column) > 0:
            return self.dat.loc[:, self.time_column].to_numpy()
        else:
            step = 1000.0 / self.sampling_rate
            final_time = self.dat.shape[0] * step
            return np.arange(0, final_time, step)

    def forces(self, session: int | None = None
    ) -> pd.DataFrame | pd.Series:
        """Returns force data of a particular column and/or a particular session"""
        columns = self.force_cols()
        if session is None:
            return self.dat.loc[:, columns]  # type: ignore
        else:
            r = self.session_range(session)
            return self.dat.loc[r.start:r.stop, columns] # type: ignore


    def add_session(self, dat: pd.DataFrame):
        """Adds data (dataframe) as a new recording

        Dataframe has to have the same columns as the already existing data
        """
        nbefore = self.dat.shape[0]
        self.dat = pd.concat([self.dat, dat], ignore_index=True)
        self.sessions.append(nbefore)


    def split_sessions(self) -> List[BeForRecord]:
        """split record in list of records of separate sessions
        """

        rtn = []
        for idx in self.session_ranges():
            dat = BeForRecord(dat=self.dat.iloc[idx, :],  # type: ignore
                        sampling_rate=self.sampling_rate,
                        time_column=self.time_column,
                        meta=self.meta)
            rtn.append(dat)

        return rtn

    def session_ranges(self) ->  List[range]:
        """list of ranges of the samples of all sessions

        """
        return [self.session_range(s) for s in range(len(self.sessions))]

    def session_range(self, session: int) -> range:
        """range of the samples (from, to) of one particular session
        """
        f = self.sessions[session]
        try:
            t = self.sessions[session + 1]
        except IndexError:
            t = self.dat.shape[0]
        return range(f, t - 1)


    def find_samples_by_time(self, times: ArrayLike) -> NDArray:
        """returns sample index (i) of the closes time in the BeForRecord.
        Takes the next larger element, if the exact time could not be found.

        .. math:: \\text{time_stamps}[i-1] <= t < \\text{time_stamps}[i]

        Parameters
        ----------
        times : ArrayLike
            the sorted array of time stamps

        """
        return np.searchsorted(self.time_stamps(), np.atleast_1d(times), "right")

    def extract_epochs(
        self,
        column: str,
        zero_samples: List[int] | NDArray[np.int_],
        n_samples: int,
        n_samples_before: int = 0,
        design: pd.DataFrame = pd.DataFrame(),
    ) -> BeForEpochs:
        """extracts epochs from BeForRecord

        Parameters
        ----------
        column: str
            name of column containing the force data to be used
        zero_samples: List[int]
            zero sample that define the epochs
        n_samples: int
            number of samples to be extract (from zero sample on)
        n_samples_before: int, optional
            number of samples to be extracted before the zero sample (default=0)

        design: pd.DataFrame, optional
            design information

        Notes
        -----
        use `find_times` to detect zero samples with time-based

        """

        fd = self.dat.loc[:, column]
        n = len(fd)  # samples for data
        n_epochs = len(zero_samples)
        n_col = n_samples_before + n_samples
        force_mtx = np.empty((n_epochs, n_col), dtype=np.float64)
        for r, zs in enumerate(zero_samples):
            f = zs - n_samples_before
            if f > 0 and f < n:
                t = zs + n_samples
                if t > n:
                    warnings.warn(
                        f"extract_force_epochs: last force epoch is incomplete, {t-n} samples missing.",
                        RuntimeWarning,
                    )
                    tmp = fd[f:]
                    force_mtx[r, : len(tmp)] = tmp
                    force_mtx[r, len(tmp) :] = 0
                else:
                    force_mtx[r, :] = fd[f:t]

        return BeForEpochs(
            force_mtx,
            sampling_rate=self.sampling_rate,
            design=design,
            zero_sample=n_samples_before,
        )



"""
Collection of useful function to deal with BeForData structures

(c) O. Lindemann
"""

import warnings as _warnings
from copy import deepcopy as _deepcopy

import numpy as _np
import pandas as _pd
from scipy import signal as _signal

from ._record import BeForRecord


def detect_sessions(
    rec: BeForRecord, time_gap: float) -> BeForRecord:
    """Detect recording sessions in the BeForRecord based on time gaps

    Parameters
    ----------
    rec : BeForRecord
        the data
    time_gap : float
        smallest time gap that should be considered as pause of the recording
        and the start of a new session

    Returns
    -------
    BeForRecord
    """

    if len(rec.time_column) == 0:
        _warnings.warn("No time column defined!", RuntimeWarning)
        return rec
    else:
        time_column = rec.time_column
    sessions = [0]
    breaks = _np.flatnonzero(_np.diff(rec.dat[time_column]) >= time_gap) + 1
    sessions.extend(breaks.tolist())
    return BeForRecord(
        rec.dat,
        sampling_rate=rec.sampling_rate,
        sessions=sessions,
        time_column=time_column,
        meta=rec.meta,
    )

def __butter_lowpass_filter(
    rec: _pd.Series, order: int, cutoff: float, sampling_rate: float, center_data: bool
):
    b, a = _signal.butter(order, cutoff, fs=sampling_rate, btype='lowpass', analog=False)  # type: ignore
    if center_data:
        # filter centred data (first sample = 0)
        return _signal.filtfilt(b, a, rec - rec.iat[0]) + rec.iat[0]
    else:
        return _signal.filtfilt(b, a, rec)


def lowpass_filter(
    rec: BeForRecord,
    cutoff: float,
    order: int,
    center_data: bool = True
) -> BeForRecord:
    """Lowpass Butterworth filter of BeforRecord

    Parameters
    ----------
    rec : BeForRecord
        the data
    cutoff : float
        cutoff frequency
    order : int
        order of the filter.
    center_data : bool (default: True)
        temporarily centred data (first sample = 0) will be used for the filtering

    Returns
    -------
    BeForRecord with filtered data

    Notes
    -----
    see documentation of `scipy.signal.butter` for information about the filtering

    """

    df = rec.dat.copy()
    for idx in rec.session_ranges():
        for c in rec.force_cols():
            df.iloc[idx, c] = __butter_lowpass_filter(rec=df.iloc[idx, c], # type: ignore
                cutoff=cutoff,
                sampling_rate=rec.sampling_rate,
                order=order,
                center_data=center_data,
            )
    meta = _deepcopy(rec.meta)
    meta["filter"] = f"butterworth: cutoff={cutoff}, order={order}"
    return BeForRecord(
        df,
        sampling_rate=rec.sampling_rate,
        sessions=rec.sessions,
        time_column=rec.time_column,
        meta=meta,
    )




"""
converting BeForData struct from and to arrow tables. Using for saving data.

(c) O. Lindemann
"""

import typing as _tp

import numpy as _np
import pandas as _pd
import pyarrow as _pa

from . import BeForEpochs, BeForRecord, _misc
from ._misc import ENC

BSL_COL_NAME = '__befor_baseline__'

def record_to_arrow(rec:BeForRecord) -> _pa.Table:
    """converts BeForRecord to ``pyarrow.Table``

    metadata of schema will be defined. Files can converted back to
    BeForRecord struct using ``BeForRecord.from_arrow()``

    Returns
    -------
    pyarrow.Table

    Examples
    --------
    >>> from pyarrow import feather
    >>> feather.write_feather(data.to_arrow(), "filename.feather",
                        compression="lz4", compression_level=6)
    """

    # Convert the DataFrame to a PyArrow table
    table = _pa.Table.from_pandas(rec.dat, preserve_index=False)

    # Add metadata to the schema (serialize sampling_rate, timestamp, trigger, and meta)
    schema_metadata = {
        "sampling_rate": str(rec.sampling_rate),
        "time_column": rec.time_column,
        "sessions": ",".join([str(x) for x in rec.sessions]),
    }
    schema_metadata.update(_misc.values_as_string(rec.meta))
    return table.replace_schema_metadata(schema_metadata)

def arrow_to_record(
    tbl: _pa.Table,
    sampling_rate: float | None = None,
    sessions: _tp.List[int] | None = None,
    time_column: str | None = None,
    meta: dict | None = None,
) -> BeForRecord:
    """Creates BeForRecord struct from `pyarrow.Table`

    Parameters
    ----------
    tbl : pyarrow.Table

    Examples
    --------
    >>> from pyarrow import feather
    >>> dat = BeForRecord.from_arrow(feather.read_table("my_force_data.feather"))

    """

    if not isinstance(tbl, _pa.Table):
        raise TypeError(f"must be pyarrow.Table, not {type(tbl)}")

    # search arrow meta data for befor record parameter
    arrow_meta = {}
    if tbl.schema.metadata is not None:
        for k, v in tbl.schema.metadata.items():
            if k == b"sampling_rate":
                if sampling_rate is None:
                    sampling_rate = _misc.try_num(v)
            elif k == b"time_column":
                if time_column is None:
                    time_column = v.decode(ENC)
            elif k == b"sessions":
                if sessions is None:
                    sessions = [int(x) for x in v.decode(ENC).split(",")]
            else:
                arrow_meta[k.decode(ENC)] = _misc.try_num(v.decode(ENC).strip())

    # make befor meta data
    if isinstance(meta, dict):
        meta.update(arrow_meta)
    else:
        meta = arrow_meta

    if sampling_rate is None:
        raise RuntimeError("No sampling rate defined!")
    if time_column is None:
        time_column = ""
    if sessions is None:
        sessions = []

    return BeForRecord(
        dat=tbl.to_pandas(),
        sampling_rate=sampling_rate,
        sessions=sessions,
        time_column=time_column,
        meta=meta
    )


def epochs_to_arrow(rec:BeForEpochs) -> _pa.Table:
    """converts BeForEpochs to ``pyarrow.Table``

    Samples and design will be concatenated to one arrow table. If baseline
    is adjusted, additionally the baseline value will be added a column.

    Zero sample and sampling_rate will be included to schema meta data.
    of schema will be defined.

    Arrow tables can converted back to BeForRecord struct using
    ``BeForEpochs.from_arrow()``

    Returns
    -------
    pyarrow.Table

    """

    dat = _pd.concat([_pd.DataFrame(rec.dat), rec.design], axis=1)
    if rec.is_baseline_adjusted():
        dat[BSL_COL_NAME] = rec.baseline
    tbl = _pa.Table.from_pandas(dat, preserve_index=False)

    schema_metadata = {
        "sampling_rate": str(rec.sampling_rate),
        "zero_sample": str(rec.zero_sample)
    }
    return tbl.replace_schema_metadata(schema_metadata)


def arrow_to_epochs(
    tbl: _pa.Table,
    sampling_rate: float | None = None,
    zero_sample: int | None = None,
) -> BeForEpochs:
    """Creates BeForEpoch struct from `pyarrow.Table`

    Parameters
    ----------
    tbl : pyarrow.Table

    """

    if not isinstance(tbl, _pa.Table):
        raise TypeError(f"must be pyarrow.Table, not {type(tbl)}")

    # search arrow meta data for befor parameter
    if tbl.schema.metadata is not None:
        for k, v in tbl.schema.metadata.items():
            if k == b"sampling_rate":
                if sampling_rate is None:
                    sampling_rate = _misc.try_num(v)
            elif k == b"zero_sample":
                if zero_sample is None:
                    try:
                        zero_sample = int(_misc.try_num(v))
                    except ValueError:
                        zero_sample = 0

    if sampling_rate is None:
        raise RuntimeError("No sampling rate defined!")
    if zero_sample is None:
        zero_sample = 0

    dat = tbl.to_pandas()

    try:
        baseline = _np.array(dat.pop(BSL_COL_NAME))
    except KeyError:
        baseline = _np.array([])

    # n_epoch_samples: count columns_name that have not int as name
    n_epoch_samples = dat.shape[1]
    for cn in reversed(dat.columns):
        try:
            int(cn)
            break
        except ValueError:
            n_epoch_samples -= 1

    return BeForEpochs(
        dat= dat.iloc[:, :n_epoch_samples],
        sampling_rate=sampling_rate,
        design=dat.iloc[:, n_epoch_samples:],
        baseline=baseline,
        zero_sample=zero_sample
    )


"""
reading xdf stream data and converts to BeForData

(c) O. Lindemann
"""

import typing as _tp

import numpy as _np
import pandas as _pd

from ._record import BeForRecord

TIME_STAMPS = "time_stamps"

def _get_channel_id(xdf_streams:_tp.List[dict], name_or_id: int | str) -> int:
    if isinstance(name_or_id, int):
        return name_or_id

    for id_, stream in enumerate(xdf_streams):
        if stream["info"]["name"][0] == name_or_id:
            return id_
    raise ValueError(f"Can't find channel {name_or_id}")

def channel_info(xdf_streams: _tp.List[dict], channel : int | str) -> _tp.Dict:
    """channel info from xdf stream data

    Args
    ----
    xdf: list of dicts
        xdf streams (as returned by `pyxdf.load_xdf`)
    channel: int | str
        channel id (int) or channel name (str)

    Returns
    -------
    dict

    """
    channel_id = _get_channel_id(xdf_streams, channel)
    info = xdf_streams[channel_id]["info"]
    fields = ("name", "type", "channel_count", "channel_format") # FIXME add clock and offset
    rtn = {k: info[k][0] for k in fields}

    tmp = xdf_streams[channel_id]["clock_times"]
    rtn["clock_times"] = ",".join([str(x) for x in tmp])
    tmp = xdf_streams[channel_id]["clock_values"]
    rtn["clock_values"] = ",".join([str(x) for x in tmp])
    return rtn

def _channel_labels(xdf_streams: _tp.List[dict], channel : int | str) -> _tp.List[str]:
    """channel labels from xdf stream data

    Args
    ----
    xdf: list of dicts
        xdf streams (as returned by `pyxdf.load_xdf`)
    channel: int | str
        channel id (int) or channel name (str)

    Returns
    -------
    List[str]

    """
    channel_id = _get_channel_id(xdf_streams, channel)
    info = xdf_streams[channel_id]["info"]
    try:
        ch_info = info["desc"][0]["channels"][0]["channel"]
    except TypeError:
        ch_info = []

    if len(ch_info) == 0:
        return [info["name"][0]]
    else:
        return [x["label"][0] for x in ch_info]

def data(xdf_streams: _tp.List[dict], channel : int | str) -> _pd.DataFrame :
    """channel pandas dataframe from xdf stream data

    Args
    ----
    xdf: list of dicts
        xdf streams (as returned by `pyxdf.load_xdf`)
    channel: int | str
        channel id (int) or channel name (str)

    Returns
    -------
    pandas.DataFrame

    """
    channel_id = _get_channel_id(xdf_streams, channel)
    lbs = [TIME_STAMPS] + _channel_labels(xdf_streams, channel_id)
    dat = _np.atleast_2d(xdf_streams[channel_id]["time_series"])
    t = _np.atleast_2d(xdf_streams[channel_id]["time_stamps"]).T
    return _pd.DataFrame(_np.hstack((t, dat)), columns=lbs)

def before_record(xdf_streams: _tp.List[dict],
                  channel : int | str,
                  sampling_rate:float) -> BeForRecord:
    """BeforeData from xdf stream data

    Args
    ----
    xdf: list of dicts
        xdf streams (as returned by `pyxdf.load_xdf`)
    channel: int | str
        channel id (int) or channel name (str)
    sampling_rate: float
        the sampling rate of the force measurements

    Returns
    -------
    BeForRecord

    """
    channel_id = _get_channel_id(xdf_streams, channel)
    return BeForRecord(dat = data(xdf_streams, channel_id),
                       sampling_rate=sampling_rate,
                       time_column=TIME_STAMPS,
                       meta=channel_info(xdf_streams, channel_id))


